input{
    width: 50px;
    height: 20px;
    border-radius : 3px;
    text-align:center;
    margin:5px 5px 5px 20px;
    background-color: antiquewhite;
    
}

button{
    width: 100px;
    height: 40px;
    background-color: #ff4777;
    border: 2px ;
    border-radius: 5px;
    font-size: medium;
    color: white;
    text-align: center;
    
}

.paralist{
    float: right;
    # width: 50%;
    height: 100%;
}
#upload_data{
    float: left;
    border-width: 2px;
    border-style: dashed;
    border-radius: 5px;
    margin: 100px, 5px,5px,5px;
    display: inline-block;
    background: #fcfccb;
    text-align: center;
}
